<?php

/**
 * @file
 * The text functions for the GD module.
 */


/**
 * Generate an image containing text with the given parameters.
 */
function drawing_gd_text_to_image($img, $cx, $cy, $text, $fontsize, $font, $color, $angle = 0, $stroke, $stroke_width, $stroke_opacity, $maximum_width = 0, $fixed_width = 0, $align = 'left') {
  // Set rotation angle.
  $q_angle = -$angle;
  while ($q_angle > 0 || $q_angle <= -90) {
    $q_angle -= $q_angle > 0 ? 90 : -90;
  }
  while ($angle < 0 || $angle >= 360) {
    $angle += $angle < 0 ? 360 : -360;
  }
  $rotation = -(floor($angle / 90) * 90);

  // Calculate rotation variables.
  $rad = deg2rad($q_angle);
  $sin = -sin($rad);
  $cos = cos($rad);

  // Perform text wrapping, if necessary.
  if ($maximum_width - 1 > 0) {
    $text = drawing_gd_text_wrap($text, $fontsize, $font, $maximum_width - 1);
  }

  // Get fixed and rotated boundry box co-ordinates.
  $bbox = drawing_gd_text_imagettfbbox($fontsize, $q_angle, $font, $text);

  // Calculate dimensions of text box.
  $text_width = sqrt(pow(abs($bbox[0] - $bbox[2]), 2) + pow(abs($bbox[1] - $bbox[3]), 2));
  $text_height = sqrt(pow(abs($bbox[0] - $bbox[6]), 2) + pow(abs($bbox[1] - $bbox[7]), 2)); // Not used.

  // Calculate dimensions of box from text box.
  $box_width = max($bbox[0], $bbox[2], $bbox[4], $bbox[6]) - min($bbox[0], $bbox[2], $bbox[4], $bbox[6]);
  $box_height = max($bbox[1], $bbox[3], $bbox[5], $bbox[7]) - min($bbox[1], $bbox[3], $bbox[5], $bbox[7]);

  // Calculate dimensions of image.
  $image_width = ($fixed_width && $maximum_width > 0)
    ? ($maximum_width - 1) * $cos + $text_height * $sin
    : $box_width;
  $image_height = ($fixed_width && $maximum_width > 0)
    ? ($maximum_width - 1) * $sin + $text_height * $cos
    : $box_height;

  // Create Image.
  $image = drawing_gd_blank($image_width + 1, $image_height + 1);
/*
  $image = imagecreatetruecolor($image_width + 1, $image_height + 1);
  $back = imagecolorallocatealpha($image, 0, 0, 0, 127);
  imagefill($image, 0, 0, $back);
*/
  // Set text alignment left
  $x = -$bbox[0];
  $y = $box_height - $bbox[3];

  if (($fixed_width && $maximum_width > 0) && $align != 'left') {
    switch ($align) {

      // Set text alignment center
      case 'center':
        $x += ($image_width - $box_width) / 2;
        $y += ($image_height - $box_height) / 2;
        break;

      // Set text alignment right
      case 'right':
        $x += $image_width - $box_width;
        $y += $image_height - $box_height;
        break;
    }
  }

  // Create the textimage.
  $fore = imagecolorallocatealpha($image, $color['r'], $color['g'], $color['b'], $color['a']);
  imagettftext($image, $fontsize, $q_angle, $x, $y, $fore, $font, $text);
  if ($rotation != 0) {
    $image = imagerotate($image, $rotation, 0);
  }

  // Add a border
  if (!empty($stroke)) {
    $image = drawing_gd_text_stroke($image, $stroke_width, $stroke, $stroke_opacity);
  }
/*
  // Preserve transparency settings.
  imagealphablending($image, TRUE);
  imagesavealpha($image, TRUE);
*/
  // Merge the images
  imagecopy($img, $image, $cx, $cy, 0, 0, $image_width,  $image_height);

  return $img;
}

function drawing_gd_text_imagettfbbox($fontsize, $angle, $font, $text) {
  // Get standard boundary box co-ordinates.
  $coords = imagettfbbox($fontsize, 0, $font, $text);

  // Get boundary box co-ordinates for firstline of text.
  $firstline_coords = $coords;
  if (strstr($text, "\n")) {
    $lines = explode("\n", $text);
    $firstline_coords = imagettfbbox($fontsize, 0, $font, $lines[0]);
  }

  // Get boundary box co-ordinates for baseline characters.
  $baseline_coords = imagettfbbox($fontsize, 0, $font, join(drupal_map_assoc(range(33, 122), 'chr')));

  // Adjust boundary box co-ordinates.
  $yoffset = ($firstline_coords[1] - $firstline_coords[5] >= $fontsize) ? $baseline_coords[1] + 1 - ($fontsize * 0.3) : 0;
  $coords[0] = 0;
  $coords[1] -= $yoffset;
  $coords[2] = $coords[2] - $baseline_coords[6];
  $coords[3] -= $yoffset;
  $coords[4] = $coords[4] - $baseline_coords[0];
  $coords[5] -= $yoffset;
  $coords[6] = 0;
  $coords[7] -= $yoffset;

  // Calculate rotation variables.
  $rad = deg2rad($angle);
  $sin = -sin($rad);
  $cos = cos($rad);

  // Rotate boundary box co-ordinates.
  $bbox = array();
  for ($i = 0; $i < 7; $i += 2) {
    $bbox[$i] = round($coords[$i] * $cos + $coords[$i + 1] * -$sin);
    $bbox[$i + 1] = round($coords[$i + 1] * $cos - $coords[$i] * -$sin);
  }

  // Return boundary box;
  return $bbox;
}

/**
 * Wrap text for rendering at a given width.
 */
function drawing_gd_text_wrap($text, $fontsize, $font, $maximum_width) {
  // State variables for the search interval
  $end = 0;
  $begin = 0;
  $fit = $begin;

  // Note: we count in bytes for speed reasons, but maintain character boundaries.
  while (true) {
    // Find the next wrap point (always after trailing whitespace).
    if (drupal_preg_match('/['. PREG_CLASS_PUNCTUATION .']['. PREG_CLASS_SEPARATOR .']*|['. PREG_CLASS_SEPARATOR .']+/u', $text, $match, PREG_OFFSET_CAPTURE, $end)) {
      $end = $match[0][1] + drupal_strlen($match[0][0]);
    }
    else {
      $end = drupal_strlen($text);
    }

    // Fetch text, removing trailing white-space and measure it.
    $line = preg_replace('/['. PREG_CLASS_SEPARATOR .']+$/u', '', drupal_substr($text, $begin, $end - $begin));
    $width = drawing_gd_text_measure_width($line, $fontsize, $font);

    // See if $line extends past the available space.
    if ($width > $maximum_width) {
      // If this is the first word, we need to truncate it.
      if ($fit == $begin) {
        // Cut off letters until it fits.
        while (drupal_strlen($line) > 0 && $width > $maximum_width) {
          $line = drupal_substr($line, 0, -1);
          $width = drawing_gd_text_measure_width($line, $fontsize, $font);
        }
        // If no fit was found, the image is too narrow..
        $fit = drupal_strlen($line) ? $begin + drupal_strlen($line) : $end;
      }

      // We have a valid fit for the next line. Insert a line-break and reset
      // the search interval.
      $text = drupal_substr($text, 0, $fit) ."\n". drupal_substr($text, $fit);
      $end = $begin = ++$fit;
    }
    else {
      // We can fit this text. Wait for now.
      $fit = $end;
    }

    if ($end == drupal_strlen($text)) {
      // All text fits. No more changes are needed.
      break;
    }
  }
  return $text;
}

/**
 * Helper function for wrapping text (measures width).
 */
function drawing_gd_text_measure_width($text, $fontsize, $font) {
  $box = imagettfbbox($fontsize, 0, $font, $text);
  return abs($box[4] - $box[0]) + 4;
}

/**
 * Stroke function adds a solid color stroke around an image with a transparent
 * background.
 *
 * @param $img
 *   The gd image resource of the image to modify
 * @param $thickness
 *   The width of the stroke to apply
 * @param $color
 *   The color of the stroke to apply
 *
 */
function drawing_gd_text_stroke($img, $thickness, $color, $text_opacity) {
  if ($thickness > 0) {
    set_time_limit(120);

    $width  = imagesx($img);
    $height = imagesy($img);

    // Create a new image which we'll lay over the original
    $border_img = drawing_gd_blank($width, $height);
    /*
    $border_img = imagecreatetruecolor($width, $height);
    $back = imagecolorallocatealpha($border_img, 0, 0, 0, 127);
    imagefill($border_img, 0, 0, $back);
    */

    for ($x = 0; $x < $width; $x++) {
      for ($y = 0; $y < $height; $y++) {
        $c = imagecolorsforindex($img, imagecolorat($img, $x, $y ));
        // Outside only modify pixels which are less opaque than the text opacity.
        if ($c['alpha'] > $text_opacity) {
          drawing_gd_text_stroke_change_pixels($img, $border_img, $thickness, $color, $x, $y, $width, $height);
        }
      }
    }
/*
    // Merge the images
    imagealphablending($img, TRUE);
*/
    imagecopy($img, $border_img, 0, 0, 0, 0, $width,  $height);
  }

  return $img;
}

/**
 * Utility function for image_stroke. Analyzes surrounding pixels and determines
 * opacity of a pixel at that x-y coordinate
 */
function drawing_gd_text_stroke_change_pixels(&$img, &$border_img, $thickness, $color, $x, $y, $width, $height) {
  extract($color);

  $pixel = imagecolorsforindex($img, imagecolorat($img, $x, $y));

  // Preform a radial analysis of all pixels within the radius of $thickness pixels
  $degree_increment = (90 / $thickness);
  $radial_coords = array();

  for ($degrees = 0; $degrees <= 90; $degrees += $degree_increment) {
    $x_offset = round(cos($degrees) * $thickness);
    $y_offset = round(sin($degrees) * $thickness);

    // Add the coordinates for the corresponding pixel in each 90° quadrant
    $radial_coords[] = array('x' => $x + $x_offset, 'y' => $y + $y_offset);
    $radial_coords[] = array('x' => $x - $x_offset, 'y' => $y + $y_offset);
    $radial_coords[] = array('x' => $x + $x_offset, 'y' => $y - $y_offset);
    $radial_coords[] = array('x' => $x - $x_offset, 'y' => $y - $y_offset);
  }

  // Generate a total alpha level for all analyzed pixels
  $total_alpha = 0;
  $total_colors = 0;

  foreach ($radial_coords as $coords) {
    if ($coords['x'] >= 0 && $coords['y'] >= 0 && $coords['x'] < $width && $coords['y'] < $height) {
      $xy_color = imagecolorsforindex($img, imagecolorat($img, $coords['x'], $coords['y']));
    }
    else {
      // This analized pixel is outside the dimensions of the image, record as transparent
      $xy_color = array('alpha' => '127');
    }

    $total_alpha += $xy_color['alpha'];
    $total_colors++;
  }

  // Check that we're not in the middle of the image or in a blonk area
  if ($total_alpha < (127 * $total_colors) && $total_alpha > 0) {
    // If we're on a semi-transparent pixel, blend the remaining amount with our border color
    if ($pixel['alpha'] < 127) {
      $alpha = 127 - $pixel['alpha'];
    }
    // We're on a completely transparent pixel where we'll use a generated transparency
    else {
      $alpha = 127 - ((127 * $total_colors) - $total_alpha);
    }
    $alpha = ($alpha < 0) ? 0 : $alpha;
    $alpha = ($alpha > 127) ? 127 : $alpha;
    // Apply the color to the border overlay image
    $color = imagecolorallocatealpha($border_img, $r, $g, $b, $alpha);
    imagesetpixel($border_img, $x, $y, $color);
  }
}